/*
 * Copyright (C) 2020 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package android.searchuiservice.cts;

import static android.searchuiservice.cts.SearchUiUtils.QUERY_INPUT;
import static android.searchuiservice.cts.SearchUiUtils.QUERY_TIMESTAMP;
import static android.searchuiservice.cts.SearchUiUtils.RESULT_CORPUS1;
import static android.searchuiservice.cts.SearchUiUtils.RESULT_CORPUS2;
import static android.searchuiservice.cts.SearchUiUtils.RESULT_CORPUS3;
import static android.searchuiservice.cts.SearchUiUtils.generateSearchTargetList;
import static android.searchuiservice.cts.SearchUiUtils.generateQuery;

import static androidx.test.InstrumentationRegistry.getContext;

import static com.google.common.truth.Truth.assertWithMessage;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.reset;
import static org.mockito.Mockito.timeout;
import static org.mockito.Mockito.verify;

import android.app.search.Query;
import android.app.search.SearchContext;
import android.app.search.SearchSession;
import android.app.search.SearchTarget;
import android.app.search.SearchTargetEvent;
import android.app.search.SearchUiManager;
import android.content.Context;
import android.os.Process;
import android.util.Log;

import androidx.annotation.NonNull;
import androidx.test.InstrumentationRegistry;
import androidx.test.runner.AndroidJUnit4;

import com.android.compatibility.common.util.RequiredServiceRule;
import com.android.compatibility.common.util.SystemUtil;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;

import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;

/**
 * Tests for {@link SearchUiManager}
 *
 * atest CtsSearchUiServiceTestCases
 */
@RunWith(AndroidJUnit4.class)
public class SearchUiManagerTest {

    private static final String TAG = "SearchUiManagerTest";
    private static final boolean DEBUG = false;

    private static final long VERIFY_TIMEOUT_MS = 5_000;
    private static final long SERVICE_LIFECYCLE_TIMEOUT_MS = 20_000;

    @Rule
    public final RequiredServiceRule mRequiredServiceRule =
            new RequiredServiceRule(Context.SEARCH_UI_SERVICE);

    private SearchUiManager mManager;
    private SearchSession mClient;
    private CtsSearchUiService.Watcher mWatcher;

    @Before
    public void setUp() throws Exception {
        mWatcher = CtsSearchUiService.setWatcher();
        mManager = getContext().getSystemService(SearchUiManager.class);
        setService(CtsSearchUiService.SERVICE_NAME);
        SearchContext searchContext = new SearchContext(
                RESULT_CORPUS1 | RESULT_CORPUS2 | RESULT_CORPUS3, 0, null);
        mClient = mManager.createSearchSession(searchContext);
        await(mWatcher.created, "Waiting for onCreate()");
        reset(mWatcher.verifier);
    }

    @After
    public void tearDown() throws Exception {
        Log.d(TAG, "Starting tear down, watcher is: " + mWatcher);
        mClient.destroy();
        setService(null);
        await(mWatcher.destroyed, "Waiting for onDestroy()");

        mWatcher = null;
        CtsSearchUiService.clearWatcher();
    }

    @Test
    public void testCreateSearchSession() {
        assertNotNull(mClient);
        assertNotNull(mWatcher.verifier);
    }

    @Test
    public void testNotifyEvent() {
        String targetId = "sample target id";
        int action = SearchTargetEvent.ACTION_SURFACE_VISIBLE;
        SearchTargetEvent event = new SearchTargetEvent.Builder(targetId, action)
                .setFlags(SearchTargetEvent.FLAG_IME_SHOWN)
                .setLaunchLocation("1,0")
                .build();
        mClient.notifyEvent(generateQuery(), event);

        ArgumentCaptor<Query> queryArg = ArgumentCaptor.forClass(Query.class);
        ArgumentCaptor<SearchTargetEvent> eventArg
                = ArgumentCaptor.forClass(SearchTargetEvent.class);

        verify(mWatcher.verifier, timeout(VERIFY_TIMEOUT_MS))
                .onNotifyEvent(any(), queryArg.capture(), eventArg.capture());

        assertTrue(queryArg.getValue().getInput().equals(QUERY_INPUT));
        assertEquals(queryArg.getValue().getTimestamp(), QUERY_TIMESTAMP);
        assertTrue(eventArg.getValue().getTargetId().equals(targetId));
        assertEquals(eventArg.getValue().getAction(), action);
    }

    @Test
    public void testQuery_realCallback() {
        Query query = SearchUiUtils.generateQuery();
        List<SearchTarget> targets = SearchUiUtils.generateSearchTargetList(3);

        final ConsumerVerifier callbackVerifier = new ConsumerVerifier(targets /* expected */);
        mWatcher.setTargets(targets /* actual */);
        mClient.query(query, Executors.newSingleThreadExecutor(), callbackVerifier);
    }

    // flaky: 8 failure out of 100
    @Ignore
    public void testQuery_mockCallback() {
        List<SearchTarget> targets = SearchUiUtils.generateSearchTargetList(2);
        Query query = SearchUiUtils.generateQuery();

        final ConsumerVerifier callbackVerifier = new ConsumerVerifier(targets);
        mClient.query(query, Executors.newSingleThreadExecutor(), callbackVerifier);

        doAnswer(answer -> {
            Consumer<List<SearchTarget>> consumer
                    = (Consumer<List<SearchTarget>>) answer.getArgument(2);
            consumer.accept(targets);
            return null;
        }).when(mWatcher.verifier).onQuery(any(), any(), any(Consumer.class));
    }

    @Test
    public void testQuery_params() {
        List<SearchTarget> targets = generateSearchTargetList(2, true, false, false, false);
        Query query = SearchUiUtils.generateQuery();

        final ConsumerVerifier callbackVerifier = new ConsumerVerifier(targets);
        mClient.query(query, Executors.newSingleThreadExecutor(), callbackVerifier);

        ArgumentCaptor<Query> queryArg = ArgumentCaptor.forClass(Query.class);
        ArgumentCaptor<Consumer<List<SearchTarget>>> callbackArg
                = ArgumentCaptor.forClass(Consumer.class);

        verify(mWatcher.verifier, timeout(VERIFY_TIMEOUT_MS))
                .onQuery(any(), queryArg.capture(), callbackArg.capture());

        Query expectedQuery = queryArg.getValue();
        assertTrue(expectedQuery.getInput().equals(QUERY_INPUT));
        assertEquals(expectedQuery.getTimestamp(), QUERY_TIMESTAMP);

        Consumer<List<SearchTarget>> expectedCallback = callbackArg.getValue();
        expectedCallback.andThen(callbackVerifier);
    }

    private void setService(String service) {
        Log.d(TAG, "Setting search ui service to " + service);
        int userId = Process.myUserHandle().getIdentifier();
        if (service != null) {
            runShellCommand("cmd search_ui set temporary-service "
                    + userId + " " + service + " 60000");
        } else {
            runShellCommand("cmd search_ui set temporary-service " + userId);
        }
    }

    private void await(@NonNull CountDownLatch latch, @NonNull String message) {
        try {
            assertWithMessage(message).that(
                    latch.await(SERVICE_LIFECYCLE_TIMEOUT_MS, TimeUnit.MILLISECONDS)).isTrue();
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            throw new IllegalStateException("Interrupted while: " + message);
        }
    }

    private void runShellCommand(String command) {
        Log.d(TAG, "runShellCommand(): " + command);
        try {
            SystemUtil.runShellCommand(InstrumentationRegistry.getInstrumentation(), command);
        } catch (Exception e) {
            throw new RuntimeException("Command '" + command + "' failed: ", e);
        }
    }

    public static class ConsumerVerifier implements
            Consumer<List<SearchTarget>> {

        private static List<SearchTarget> mExpectedTargets;

        public ConsumerVerifier(List<SearchTarget> targets) {
            mExpectedTargets = targets;
        }

        @Override
        public void accept(List<SearchTarget> actualTargets) {
            if (DEBUG) {
                Log.d(TAG, "ConsumerVerifier.accept targets.size= " + actualTargets.size());
                Log.d(TAG, "ConsumerVerifier.accept target(1).packageName=" + actualTargets.get(
                        0).getPackageName());
            }
            Assert.assertArrayEquals(actualTargets.toArray(), mExpectedTargets.toArray());
        }
    }
}
